import sys
from datetime import datetime
import pytz
from performanceplatform.collector.piwik.base import BaseFetcher
from performanceplatform.utils.data_pusher import Pusher


class Fetcher(BaseFetcher):
    def __init__(self, credentials, query):
        super(Fetcher, self).__init__(credentials, query)
        self.last_minutes = query.get('minutes', 2)

    def _request_params(self):
        params = super(Fetcher, self)._request_params()
        params.update({'lastMinutes': self.last_minutes})
        return params


class Parser():
    def _timestamp(self):
        timezone = pytz.UTC
        timestamp = datetime.now(timezone).replace(microsecond=0)
        return timestamp.isoformat()

    def parse(self, data):
        timestamp = self._timestamp()
        visitor_count = int(data[0]["visitors"])
        return [{
            "_timestamp": timestamp,
            "_id": timestamp,
            "unique_visitors": visitor_count,
            "for_url": ""}]


def main(credentials, data_set_config, query, options, start_at, end_at):
    if start_at or end_at:
        print 'Cannot backfill realtime collectors'
        sys.exit(1)

    data = Fetcher(credentials, query).fetch()

    parsed_data = Parser().parse(data)

    Pusher(data_set_config, options).push(parsed_data)



import logging
import time
from performanceplatform.collector.logging_setup import (
    extra_fields_from_exception)

# import everything from requests so that other code can import this package
# instead of requests
from requests import *

_MAX_RETRIES = 5


def get(url, *args, **kwargs):
    return __request_with_backoff('GET', url, *args, **kwargs)


def post(url, *args, **kwargs):
    return __request_with_backoff('POST', url, *args, **kwargs)


def __request_with_backoff(method, url, *args, **kwargs):
    delay = 10

    for n in range(_MAX_RETRIES):
        response = request(method, url, *args, **kwargs)
        code = response.status_code
        if code in [403, 502, 503]:
            retry_info = ('{} request for {} failed with code {} '
                          '(Attempt {} of {}).'.format(method, url, code,
                                                       n + 1, _MAX_RETRIES))
            if n + 1 < _MAX_RETRIES:
                retry_info += ' Retrying in {} seconds...'.format(delay)
            logging.info(retry_info)
            time.sleep(delay)
            delay *= 2
        else:
            response.raise_for_status()
            return response

    # we made _MAX_RETRIES requests but none worked
    logging.error('Max retries exceeded for {}'.format(url))
    try:
        response.raise_for_status()
    except Exception as e:
        logging.error('Requests Error: {}'.format(
            e.message,
            extra=extra_fields_from_exception(e)
        ))
        raise



from datetime import datetime
import pytz
import sys

from performanceplatform.collector.ga.lib.helper import create_client

from performanceplatform.client import DataSet


GOOGLE_API_SCOPE = "https://www.googleapis.com/auth/analytics"


class Collector(object):
    def __init__(self, credentials):
        self._realtime = Realtime(credentials)

    def send_records_for(self, query, to):
        data_set = DataSet.from_config(to)

        visitor_count = self._realtime.query(query)

        record = self._create_record(visitor_count,
                                     query.get('filters', ''))

        data_set.post(record)

    def _create_record(self, visitor_count, for_url):
        timestamp = _timestamp()
        return {
            "_timestamp": timestamp,
            "_id": timestamp,
            "unique_visitors": visitor_count,
            "for_url": for_url
        }


class Realtime(object):
    def __init__(self, credentials):
        self._client = create_client(credentials)

    def execute_ga_query(self, query):
        return self._client._service.data().realtime().get(
            **query
        ).execute()

    def query(self, query):
        response = self.execute_ga_query(query)

        if "rows" in response:
            visitor_count = int(response["rows"][0][0])
        else:
            visitor_count = 0

        return visitor_count


def _timestamp():
    timezone = pytz.UTC
    timestamp = datetime.now(timezone).replace(microsecond=0)
    return timestamp.isoformat()


def main(credentials, data_set, query, options, start_at, end_at):
    if start_at or end_at:
        print 'Cannot backfill realtime collectors'
        sys.exit(1)
    collector = Collector(credentials)
    collector.send_records_for(query, data_set)



from performanceplatform.client.collector import CollectorAPI


def get_config(collector_slug, performanceplatform):
    collector_client = CollectorAPI(
        performanceplatform['stagecraft_url'],
        performanceplatform['omniscient_api_token']
    )
    collector = collector_client.get_collector(collector_slug)
    return {
        'query': collector.get('query'),
        'options': collector.get('options'),
        'data-set': {
            'data-group': collector.get('data_set').get('data_group'),
            'data-type': collector.get('data_set').get('data_type'),
            'bearer_token': collector.get('data_set').get('bearer_token')
        },
        'entrypoint': collector.get('entry_point'),
        'token': collector.get('provider').get('slug')
    }



from performanceplatform.utils import requests_with_backoff
from datetime import datetime, timedelta
import pytz
from performanceplatform.collector.webtrends.base import(
    BaseCollector, BaseParser)


class Collector(BaseCollector):
    def __init__(self, credentials, query, start_at, end_at):
        self.api_version = credentials.get('api_version')
        self.base_url = credentials['reports_url']
        self.report_id = query.pop('report_id')
        super(Collector, self).__init__(credentials, query, start_at, end_at)

    @classmethod
    def date_range_for_webtrends(cls, start_at=None, end_at=None):
        """
        Get the day dates in between start and end formatted for query.
        This returns dates inclusive e.g. final day is (end_at, end_at+1 day)
        """
        if start_at and end_at:
            start_date = cls.parse_standard_date_string_to_date(
                start_at)
            end_date = cls.parse_standard_date_string_to_date(
                end_at)
            numdays = (end_date - start_date).days + 1
            start_dates = [end_date - timedelta(days=x)
                           for x in reversed(range(0, numdays))]
            date_range = []
            for i, date in enumerate(start_dates):
                query_date = cls.parse_date_for_query(date)
                date_range.append((query_date, query_date))
            return date_range
        else:
            return [("current_day-1", "current_day-1")]

    def _make_request(self, start_at_for_webtrends, end_at_for_webtrends):
        return requests_with_backoff.get(
            url="{base_url}{report_id}".format(
                base_url=self.base_url,
                report_id=self.report_id),
            auth=(self.user, self.password),
            params={
                'start_period': start_at_for_webtrends,
                'end_period': end_at_for_webtrends,
                'format': self.query_format
            }
        )

    def build_parser(self, data_set_config, options):
        if self.api_version == 'v3':
            return V3Parser(options, data_set_config['data-type'])
        else:
            return V2Parser(options, data_set_config['data-type'])


class V2Parser(BaseParser):
    def __init__(self, options, data_type):
        self.row_type_name = options['row_type_name']
        super(V2Parser, self).__init__(options, data_type)

    def parse_item(self, item):
        initial_data = []
        special_fields = []
        for date, data_point in item.items():
            start_date = self.__class__.to_datetime(date)
            if data_point['SubRows']:
                for row_type, row in data_point['SubRows'].items():
                    initial_data.append({
                        'start_date': start_date,
                    })
                    special_fields.append(dict({
                        self.row_type_name: row_type
                    }.items() + row['measures'].items()))
        return initial_data, special_fields

    @classmethod
    def to_datetime(cls, webtrends_formatted_date):
        start_of_period = webtrends_formatted_date.split("-")[0]
        return datetime.strptime(
            start_of_period, "%m/%d/%Y").replace(tzinfo=pytz.UTC)


class V3Parser(BaseParser):
    def __init__(self, options, data_type):
        self.row_type_name = options['row_type_name']
        super(V3Parser, self).__init__(options, data_type)

    def parse_item(self, items):
        initial_data = []
        special_fields = []
        for value in items:
            start_date = self.__class__.to_datetime(value)
            if value['SubRows']:
                for row_type, row in value['SubRows'][0].items():
                    initial_data.append({
                        'start_date': start_date,
                    })
                    special_fields.append(dict({
                        self.row_type_name: row_type
                    }.items() + row['measures'].items()))
        return initial_data, special_fields

    @classmethod
    def to_datetime(cls, webtrends_formatted_data_point):
        start_at = webtrends_formatted_data_point['start_date']
        date_parts = start_at.split('-')
        if len(date_parts) == 3:
            return datetime.strptime(
                start_at, "%Y-%m-%d").replace(tzinfo=pytz.UTC)
        else:
            return datetime.strptime(
                start_at, "%Y-%m").replace(tzinfo=pytz.UTC)


def main(credentials, data_set_config, query, options, start_at, end_at):
    collector = Collector(credentials, query, start_at, end_at)
    collector.collect_parse_and_push(data_set_config, options)



from logstash_formatter import LogstashFormatter
import logging
import os
import sys
import traceback


def get_log_file_handler(path):
    handler = logging.FileHandler(path)
    handler.setFormatter(logging.Formatter(
        "%(asctime)s [%(levelname)s] -> %(message)s"))
    return handler


def get_json_log_handler(path, app_name, json_fields):
    handler = logging.FileHandler(path)
    formatter = LogstashFormatter()
    formatter.defaults['@tags'] = ['collector', app_name]
    formatter.defaults['@fields'] = json_fields
    handler.setFormatter(formatter)
    return handler


def uncaught_exception_handler(*exc_info):
    text = "".join(traceback.format_exception(*exc_info))
    (exc_class, exc_message, _) = exc_info
    extra = {
        'exception_class': exc_class.__name__,
        'exception_message': exc_message
    }
    logging.error("Unhandled exception: %s", text, extra=extra)


def set_up_logging(
        app_name,
        log_level,
        logfile_path,
        logfile_name,
        json_fields=None):

    if logfile_name is None:
        logfile_name = 'production'
    sys.excepthook = uncaught_exception_handler
    logger = logging.getLogger()
    logger.setLevel(log_level)
    logger.addHandler(get_log_file_handler(
        os.path.join(logfile_path, '{}.log'.format(logfile_name))))
    logger.addHandler(get_json_log_handler(
        os.path.join(logfile_path, '{}.json.log'.format(logfile_name)),
        app_name,
        json_fields=json_fields if json_fields else {}))
    logger.info("{0} logging started".format(app_name))


def extra_fields_from_exception(exception):
    """Exception into a dict to be passed to logger

    Standardise the way we send exception messages so they can be filtered in
    kibana, also make it easy to add extra information in future.
    """
    return {
        'exception_class': type(exception).__name__,
        'exception_message': exception.message
    }



from performanceplatform.utils.datetimeutil \
    import to_datetime, to_utc
import logging
import base64
import re


class DataParser(object):
    def __init__(self, data, options, data_type):
        # it would be nice to have some json schemas or something to validate
        # for now here are some docs
        """
        data: Can be any array of dicts
        options: {
          mappings: "a dict: converts keys matching key of the mappings dict
              to the value in the mapping dict",
          additionalFields: "a dict:
              key value pairs will be added into the returned data",
          idMapping: "a list of keys or single key string: for each key
              the corresponding values will be concatenated in order to create
              a unique _humanId field on the returned data and base64
              encoded in order to create an _id field. If no idMapping
              are provided then the item start_date and any avaiable
              'dimensions' will be used instead",
          dataType: "a value to be set on a dataType attribute.
              Overrides the data_type argument"
          plugins: "a list of string respresentations of python classes being
              instantiated. e.g. ComputeIdFrom('_timestamp', '_timespan').
              To be run on the passed in data to mutate it.
              See performanceplatform.collector.ga.plugins for more details"
        }
        data_type: "a string - the data_type to be set as a data_type
            attribute on the returned data unless it is overridden in options"
        """
        self.data = list(data)
        if "plugins" in options:
            self.plugins = options["plugins"]
        else:
            self.plugins = None
        self.data_type = options.get('dataType', data_type)
        self.mappings = options.get("mappings", {})
        self.idMapping = options.get("idMapping", None)
        self.additionalFields = options.get('additionalFields', {})

    def get_data(self, special_fields=None):
        """
        special_fields: "a dict of data specific to collector type
            that should be added to the data returned by the parser.
            This will also be and operated on by idMapping, mappings and
            plugins"

        This method loops through the data provided to the instance.
        For each item it will return a dict of the format:
        {
            "_timestamp": "the item start_date",
            "dataType": "self.data_type for the instance",
            ...
            "any additional fields": "from self.additionalFields",
            "any special_fields": "from special fields argument",
            "any item dimensions": "from item.dimensions",
            ...
            mappings changing keys in this dict from self.mappings
               are then applied on the above
            ...
            "_humanId": "derived from either the values corresponding to
               idMapping concatenated or the data_type, item.start_date,
               timeSpan and item.dimensions values if any concatenated"
            "_id": "derived from either the values corresponding to
               idMapping concatenated or the data_type, item.start_date,
               timeSpan and item.dimensions values if any concatenated and
               then base64 encoded"
        }
        """
        docs = build_document_set(self.data, self.data_type, self.mappings,
                                  special_fields,
                                  self.idMapping,
                                  additionalFields=self.additionalFields)

        if self.plugins:
            docs = run_plugins(self.plugins, list(docs))

        return docs


def build_document_set(results, data_type, mappings, special_fields,
                       idMapping=None,
                       additionalFields={}):
    if special_fields and len(results) != len(special_fields):
        raise ValueError(
            "There must be same number of special fields as results")
    parsed = []
    for i, item in enumerate(results):
        if not special_fields:
            special = {}
        else:
            special = special_fields[i]
        parsed.append(build_document(item,
                                     data_type,
                                     special,
                                     mappings,
                                     idMapping,
                                     additionalFields=additionalFields))
    return parsed


def build_document(item, data_type, special_fields={},
                   mappings=None, idMapping=None,
                   additionalFields={}):
    if data_type is None:
        raise ValueError("Must provide a data type")
    if mappings is None:
        mappings = {}

    base_properties = {
        "_timestamp": to_datetime(item["start_date"]),
        "dataType": data_type
    }

    doc = dict(base_properties.items() +
               additionalFields.items() +
               item.get("dimensions", {}).items() +
               special_fields.items())
    doc = apply_key_mapping(mappings, doc)

    def get_value_for_key(key):
        return unicode(doc.get(key, ""))

    if idMapping is not None:
        if not isinstance(idMapping, list):
            idMapping = [idMapping]

        values_for_id = map(get_value_for_key, idMapping)
        string_for_id = "".join(values_for_id)
    else:
        string_for_id = get_string_for_data_id(
            data_type,
            to_datetime(item["start_date"]),
            doc.get('timeSpan', None),
            item.get('dimensions', {}).values())
    (_id, human_id) = value_id(string_for_id)

    doc['humanId'] = human_id
    doc['_id'] = _id

    return doc


def get_string_for_data_id(data_type, timestamp, period, dimension_values):
    # `dimension_values` may be non-string python types and need to be coerced.
    values = map(unicode, dimension_values)
    slugs = [data_type, _format(timestamp), period] + values
    slugs = [value for value in slugs
             if value is not None]
    return "_".join(slugs)


def _format(timestamp):
    return to_utc(timestamp).strftime("%Y%m%d%H%M%S")


def value_id(value):
    value_bytes = value.encode('utf-8')
    logging.debug(u"'{0}' ({1})".format(value, type(value)))
    return base64.urlsafe_b64encode(value_bytes), value_bytes


def run_plugins(plugins_strings, results):

    last_plugin = plugins_strings[-1]
    if not last_plugin.startswith("ComputeIdFrom"):
        raise RuntimeError("Last plugin must be `ComputeIdFrom` for now. This "
                           "may be changed with further development if "
                           "necessary")

    # Import is here so that it is only required when "plugins" is specified
    from performanceplatform.collector.ga.plugins import load_plugins
    plugins = load_plugins(plugins_strings)

    # Plugins are designed so that their configuration is described in the
    # plugin string. At this point, `plugin` should be a closure which only
    # requires the documents to process.
    for plugin in plugins:
        results = plugin(results)

    return results


def apply_key_mapping(mapping, pairs):
    return dict(map_one_to_one_fields(mapping, pairs).items() +
                map_multi_value_fields(mapping, pairs).items())


def map_one_to_one_fields(mapping, pairs):
    return dict((mapping.get(key, key), value) for key, value in pairs.items())


def map_multi_value_fields(mapping, pairs):
    multi_value_regexp = '^(.*)_(\d*)$'
    multi_value_delimiter = ':'
    mapped_pairs = {}

    for from_key, to_key in mapping.items():
        multi_value_matches = re.search(multi_value_regexp, from_key)
        if multi_value_matches:
            key = multi_value_matches.group(1)
            index = int(multi_value_matches.group(2))
            multi_value = pairs.get(key)
            if multi_value is None:
                continue

            values = multi_value.split(multi_value_delimiter)
            if index < len(values):
                mapped_pairs[to_key] = values[index]

    return mapped_pairs
